/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// view.c -- player eye positioning

#include "quakedef.h"
#include <time.h>	// cl_clock

/*

The view is allowed to move slightly from it's true position for bobbing,
but if it exceeds 8 pixels linear distance (spherical, not box), the list of
entities sent from the server may not include everything in the pvs, especially
when crossing a water boudnary.

*/

cvar_t	lcd_x = {"lcd_x","0"};
cvar_t	lcd_yaw = {"lcd_yaw","0"};

cvar_t	scr_ofsx = {"scr_ofsx","0", false};
cvar_t	scr_ofsy = {"scr_ofsy","0", false};
cvar_t	scr_ofsz = {"scr_ofsz","0", false};

cvar_t	scr_clock = {"cl_clock", "0"};
cvar_t	scr_clock_x = {"cl_clock_x", "0"};
cvar_t	scr_clock_y = {"cl_clock_y", "-1"};
cvar_t	show_speed = {"show_speed", "0"};
cvar_t	show_fps = {"show_fps","0"};

cvar_t	cl_rollspeed = {"cl_rollspeed", "200"};
cvar_t	cl_rollangle = {"cl_rollangle", "2.0"};

cvar_t	cl_bob = {"cl_bob","0.02", false};
cvar_t	cl_bobcycle = {"cl_bobcycle","0.6", false};
cvar_t	cl_bobup = {"cl_bobup","0.5", false};

cvar_t	v_kicktime = {"v_kicktime", "0.5", false};
cvar_t	v_kickroll = {"v_kickroll", "0.6", false};
cvar_t	v_kickpitch = {"v_kickpitch", "0.6", false};
cvar_t	v_gunkick = {"v_gunkick", "0"};		// by joe

cvar_t	v_iyaw_cycle = {"v_iyaw_cycle", "2", false};
cvar_t	v_iroll_cycle = {"v_iroll_cycle", "0.5", false};
cvar_t	v_ipitch_cycle = {"v_ipitch_cycle", "1", false};
cvar_t	v_iyaw_level = {"v_iyaw_level", "0.3", false};
cvar_t	v_iroll_level = {"v_iroll_level", "0.1", false};
cvar_t	v_ipitch_level = {"v_ipitch_level", "0.3", false};

cvar_t	v_idlescale = {"v_idlescale", "0", false};

cvar_t	crosshair = {"crosshair", "0", true};
cvar_t	crosshaircolor = {"crosshaircolor", "79", true};
cvar_t	crosshairsize	= {"crosshairsize", "1"};
cvar_t	cl_crossx = {"cl_crossx", "0", false};
cvar_t	cl_crossy = {"cl_crossy", "0", false};

cvar_t	v_damagecshift = {"v_damagecshift", "1"};
cvar_t	v_quadcshift = {"v_quadcshift", "1"};
cvar_t	v_suitcshift = {"v_suitcshift", "1"};
cvar_t	v_ringcshift = {"v_ringcshift", "1"};
cvar_t	v_pentcshift = {"v_pentcshift", "1"};

#ifdef GLQUAKE
cvar_t	v_dlightcshift = {"v_dlightcshift", "1"};
#endif

cvar_t	v_bonusflash = {"cl_bonusflash", "1"};

float	v_dmg_time, v_dmg_roll, v_dmg_pitch;

/*
===============
V_CalcRoll

Used by view and sv_user
===============
*/
vec3_t	forward, right, up;

float V_CalcRoll (vec3_t angles, vec3_t velocity)
{
	float	sign;
	float	side;
	float	value;
	
	AngleVectors (angles, forward, right, up);
	side = DotProduct (velocity, right);
	sign = side < 0 ? -1 : 1;
	side = fabs(side);
	
	value = cl_rollangle.value;

	if (side < cl_rollspeed.value)
		side = side * value / cl_rollspeed.value;
	else
		side = value;
	
	return side*sign;
	
}


/*
===============
V_CalcBob

===============
*/
float V_CalcBob (void)
{
	static	float	bob;
	float			cycle;
	
	if (cl_bobcycle.value <= 0)
		return 0;

	cycle = cl.time - (int)(cl.time / cl_bobcycle.value) * cl_bobcycle.value;
	cycle /= cl_bobcycle.value;
	if (cycle < cl_bobup.value)
		cycle = M_PI * cycle / cl_bobup.value;
	else
		cycle = M_PI + M_PI * (cycle - cl_bobup.value) / (1.0 - cl_bobup.value);

// bob is proportional to velocity in the xy plane
// (don't count Z, or jumping messes it up)
	bob = sqrt(cl.velocity[0]*cl.velocity[0] + cl.velocity[1]*cl.velocity[1]) * cl_bob.value;
	bob = bob * 0.3 + bob * 0.7 * sin(cycle);
	bob = bound(-7, bob, 4);

	return bob;
}


//=============================================================================


cvar_t	v_centermove = {"v_centermove", "0.15", false};
cvar_t	v_centerspeed = {"v_centerspeed","500"};


void V_StartPitchDrift (void)
{
#if 1
	if (cl.laststop == cl.time)
	{
		return;		// something else is keeping it from drifting
	}
#endif
	if (cl.nodrift || !cl.pitchvel)
	{
		cl.pitchvel = v_centerspeed.value;
		cl.nodrift = false;
		cl.driftmove = 0;
	}
}

void V_StopPitchDrift (void)
{
	cl.laststop = cl.time;
	cl.nodrift = true;
	cl.pitchvel = 0;
}

/*
===============
V_DriftPitch

Moves the client pitch angle towards cl.idealpitch sent by the server.

If the user is adjusting pitch manually, either with lookup/lookdown,
mlook and mouse, or klook and keyboard, pitch drifting is constantly stopped.

Drifting is enabled when the center view key is hit, mlook is released and
lookspring is non 0, or when 
===============
*/
void V_DriftPitch (void)
{
	float		delta, move;

	if (noclip_anglehack || !cl.onground || cls.demoplayback)
	{
		cl.driftmove = 0;
		cl.pitchvel = 0;
		return;
	}

// don't count small mouse motion
	if (cl.nodrift)
	{
		if (fabs(cl.cmd.forwardmove) < cl_forwardspeed.value)
			cl.driftmove = 0;
		else
			cl.driftmove += host_frametime;
	
		if (cl.driftmove > v_centermove.value)
			V_StartPitchDrift ();

		return;
	}
	
	delta = cl.idealpitch - cl.viewangles[PITCH];

	if (!delta)
	{
		cl.pitchvel = 0;
		return;
	}

	move = host_frametime * cl.pitchvel;
	cl.pitchvel += host_frametime * v_centerspeed.value;
	
//Con_Printf ("move: %f (%f)\n", move, host_frametime);

	if (delta > 0)
	{
		if (move > delta)
		{
			cl.pitchvel = 0;
			move = delta;
		}
		cl.viewangles[PITCH] += move;
	}
	else if (delta < 0)
	{
		if (move > -delta)
		{
			cl.pitchvel = 0;
			move = -delta;
		}
		cl.viewangles[PITCH] -= move;
	}
}



/*
============================================================================== 
 
				PALETTE FLASHES 
 
============================================================================== 
*/ 
 
cshift_t	cshift_empty = {{130, 80, 50}, 0};
cshift_t	cshift_water = {{130, 80, 50}, 128};
cshift_t	cshift_slime = {{0, 25, 5}, 150};
cshift_t	cshift_lava = {{255, 80, 0}, 150};

#ifdef	GLQUAKE
float			v_blend[4];	
unsigned short	v_gammaramps[3][256];
cvar_t			gl_hwblend = {"gl_hwblend", "1"};
cvar_t			gl_gamma = {"gl_gamma", "1", true};
cvar_t			gl_contrast = {"gl_contrast", "1", true};
#else
byte			gammatable[256];	// palette is sent through this
byte			current_pal[768];	// Tonik: used for screenshots
#endif

// NOTE: "gamma" is present in default.cfg, so I just left it in...
cvar_t		v_gamma = {"gamma", "1", true};
cvar_t		v_contrast = {"contrast", "1", true};


#ifndef GLQUAKE
void V_BuildGammaTable (float g, float c)
{
	int	i, inf;
	
	g = bound(0.3, g, 3);
	c = bound(1, c, 3);

	if (g == 1 && c == 1)
	{
		for (i=0 ; i<256 ; i++)
		{
			gammatable[i] = i;
		}
		return;
	}
	
	for (i=0 ; i<256 ; i++)
	{
		inf = 255 * pow ((i+0.5)/255.5*c, g) + 0.5;
		if (inf < 0)
			inf = 0;
		if (inf > 255)
			inf = 255;
		gammatable[i] = inf;
	}
}

/*
=================
V_CheckGamma
=================
*/
qboolean V_CheckGamma (void)
{
	static	float	old_gamma, old_contrast;
	
	if (v_gamma.value == old_gamma && v_contrast.value == old_contrast)
		return false;
	old_gamma = v_gamma.value;
	old_contrast = v_contrast.value;
	
	V_BuildGammaTable (v_gamma.value, v_contrast.value);
	vid.recalc_refdef = 1;			// force a surface cache flush
	
	return true;
}
#endif // GLQUAKE

/*
===============
V_ParseDamage
===============
*/
void V_ParseDamage (void)
{
	int			i, armor, blood;
	vec3_t		from, forward, right, up;
	entity_t	*ent;
	float		side, count, fraction;
	
	armor = MSG_ReadByte ();
	blood = MSG_ReadByte ();
	for (i=0 ; i<3 ; i++)
		from[i] = MSG_ReadCoord ();

	count = blood*0.5 + armor*0.5;
	if (count < 10)
		count = 10;

	cl.faceanimtime = cl.time + 0.2;		// but sbar face into pain frame

	cl.cshifts[CSHIFT_DAMAGE].percent += 3*count;
	if (cl.cshifts[CSHIFT_DAMAGE].percent < 0)
		cl.cshifts[CSHIFT_DAMAGE].percent = 0;
	if (cl.cshifts[CSHIFT_DAMAGE].percent > 150)
		cl.cshifts[CSHIFT_DAMAGE].percent = 150;

	fraction = bound(0, v_damagecshift.value, 1);
	cl.cshifts[CSHIFT_DAMAGE].percent *= fraction;

	if (armor > blood)		
	{
		cl.cshifts[CSHIFT_DAMAGE].destcolor[0] = 200;
		cl.cshifts[CSHIFT_DAMAGE].destcolor[1] = 100;
		cl.cshifts[CSHIFT_DAMAGE].destcolor[2] = 100;
	}
	else if (armor)
	{
		cl.cshifts[CSHIFT_DAMAGE].destcolor[0] = 220;
		cl.cshifts[CSHIFT_DAMAGE].destcolor[1] = 50;
		cl.cshifts[CSHIFT_DAMAGE].destcolor[2] = 50;
	}
	else
	{
		cl.cshifts[CSHIFT_DAMAGE].destcolor[0] = 255;
		cl.cshifts[CSHIFT_DAMAGE].destcolor[1] = 0;
		cl.cshifts[CSHIFT_DAMAGE].destcolor[2] = 0;
	}

	// calculate view angle kicks
	ent = &cl_entities[cl.viewentity];
	
	VectorSubtract (from, ent->origin, from);
	VectorNormalize (from);
	
	AngleVectors (ent->angles, forward, right, up);

	side = DotProduct (from, right);
	v_dmg_roll = count*side*v_kickroll.value;
	
	side = DotProduct (from, forward);
	v_dmg_pitch = count*side*v_kickpitch.value;

	v_dmg_time = v_kicktime.value;
}


/*
==================
V_cshift_f
==================
*/
void V_cshift_f (void)
{
	cshift_empty.destcolor[0] = atoi(Cmd_Argv(1));
	cshift_empty.destcolor[1] = atoi(Cmd_Argv(2));
	cshift_empty.destcolor[2] = atoi(Cmd_Argv(3));
	cshift_empty.percent = atoi(Cmd_Argv(4));
}


/*
==================
V_BonusFlash_f

When you run over an item, the server sends this command
==================
*/
void V_BonusFlash_f (void)
{
	if (!v_bonusflash.value) return;

	cl.cshifts[CSHIFT_BONUS].destcolor[0] = 215;
	cl.cshifts[CSHIFT_BONUS].destcolor[1] = 186;
	cl.cshifts[CSHIFT_BONUS].destcolor[2] = 69;
	cl.cshifts[CSHIFT_BONUS].percent = 50;
}

/*
=============
V_SetContentsColor

Underwater, lava, etc each has a color shift
=============
*/
void V_SetContentsColor (int contents)
{
	cshift_t* c;

	c = &cl.cshifts[CSHIFT_CONTENTS]; // just to shorten the code below

	if (!gl_polyblend.value)
	{
		c->percent = 0;
		return;
	}
	switch (contents)
	{
	case CONTENTS_EMPTY:
	case CONTENTS_SOLID:
		c->destcolor[0] = cshift_empty.destcolor[0];
		c->destcolor[1] = cshift_empty.destcolor[1];
		c->destcolor[2] = cshift_empty.destcolor[2];
		c->percent = cshift_empty.percent;
		break;
	case CONTENTS_LAVA:
		c->destcolor[0] = cshift_lava.destcolor[0];
		c->destcolor[1] = cshift_lava.destcolor[1];
		c->destcolor[2] = cshift_lava.destcolor[2];
		c->percent = cshift_lava.percent;
		break;
	case CONTENTS_SLIME:
		c->destcolor[0] = cshift_slime.destcolor[0];
		c->destcolor[1] = cshift_slime.destcolor[1];
		c->destcolor[2] = cshift_slime.destcolor[2];
		c->percent = cshift_slime.percent;
		break;
	default:
		c->destcolor[0] = cshift_water.destcolor[0];
		c->destcolor[1] = cshift_water.destcolor[1];
		c->destcolor[2] = cshift_water.destcolor[2];
		c->percent = cshift_water.percent;
		break;
	}
}


#ifdef GLQUAKE
// totally utterly overkill but if it works :)
/*
=============
V_Setfog

Sets fog based on contents
=============
*/
void V_Setfog (int contents)
{
	if (ISINMAP(contents) || !ISUNDERWATER(contents)) {
		V_AddMapfog (contents);
	} else if (ISUNDERWATER(contents) || ISINMAP(contents)) {
		V_AddWaterfog (contents);
	}
}

/*
=============
V_AddWaterFog

Fog in liquids, from FuhQuake
=============
*/
void V_AddWaterfog (int contents)
{
	GLfloat	vcolors[4];
	GLfloat	lava[4] = {(GLfloat)1.0f, (GLfloat)0.314f, (GLfloat)0.0f, (GLfloat)0.5f};
	GLfloat	slime[4] = {(GLfloat)0.039f, (GLfloat)0.738f, (GLfloat)0.333f, (GLfloat)0.5f};
	GLfloat	water[4] = {(GLfloat)0.250f, (GLfloat)0.188f, (GLfloat)0.0f, (GLfloat)0.5f};
	vec3_t  diff;
	float   dist;

	if (!gl_waterfog.value) {
		glDisable(GL_FOG);
		r_viewleaf->waterfogenabled = false;
		return;
	}
	VectorSubtract(currententity->origin, r_origin, diff);

	dist = VectorLength (diff);

	contents = CL_PointContents(r_origin);

	switch (contents)
	{
	case CONTENTS_LAVA:
		glEnable(GL_FOG);
		r_viewleaf->waterfogenabled = true;
		vcolors[0] = lava[0];
		vcolors[1] = lava[1];
		vcolors[2] = lava[2];
		vcolors[3] = exp(gl_waterfog_density.value / dist);
		break;

	case CONTENTS_SLIME:
		glEnable(GL_FOG);
		r_viewleaf->waterfogenabled = true;
		vcolors[0] = slime[0];
		vcolors[1] = slime[1];
		vcolors[2] = slime[2];
		vcolors[3] = exp(gl_waterfog_density.value / dist);
		break;

	case CONTENTS_WATER:
		glEnable(GL_FOG);
		r_viewleaf->waterfogenabled = true;
		vcolors[0] = water[0];
		vcolors[1] = water[1];
		vcolors[2] = water[2];
		vcolors[3] = exp(gl_waterfog_density.value / dist);
		break;
		
	default:
		glDisable(GL_FOG);
		r_viewleaf->waterfogenabled = false;
		vcolors[0] = 0.0f;
		vcolors[1] = 0.0f;
		vcolors[2] = 0.0f;
		vcolors[3] = 0.0f;
		break;
	}
	glFogfv (GL_FOG_COLOR, vcolors);

	if (r_viewleaf->waterfogenabled) {
		if (((int)gl_waterfog.value) == 2) {
			glFogf (GL_FOG_DENSITY, 0.0002 + (0.0009 - 0.0002) * bound(0.0f, (GLfloat)gl_waterfog_density.value, 1.0f));
			glFogi (GL_FOG_MODE, GL_EXP);
		} else {
			glFogi (GL_FOG_MODE, GL_LINEAR);
			glFogf (GL_FOG_START, 150.0f);	
			glFogf (GL_FOG_END, 4250.0f - (4250.0f - 1536.0f) * bound(0.0f, (GLfloat)gl_waterfog_density.value, 1.0f));
		}
	}
}

/*
=============
V_AddMapFog

Fog in World
=============
*/
void V_AddMapfog (int contents)
{
	vec4_t	mcolors;
	vec3_t  diff;
	float   dist;

	if (!gl_fogenable.value) {
		glDisable(GL_FOG);
		r_viewleaf->worldfogenabled = false;
		return; // arrghh 
	}
	VectorSubtract(currententity->origin, r_origin, diff);

	dist = VectorLength (diff);

	contents = CL_PointContents(r_origin);

	switch (contents)
	{
	case CONTENTS_SOLID:
		glEnable(GL_FOG);
		r_viewleaf->worldfogenabled = true;
		mcolors[0] = bound(0.0f, gl_fogred.value, 1.0f);
		mcolors[1] = bound(0.0f, gl_foggreen.value, 1.0f);
		mcolors[2] = bound(0.0f, gl_fogblue.value, 1.0f); 
		mcolors[3] = exp(gl_fogdensity.value / dist);
		break;

	case CONTENTS_EMPTY:
		glEnable(GL_FOG);
		r_viewleaf->worldfogenabled = true;
		mcolors[0] = bound(0.0f, gl_fogred.value, 1.0f);
		mcolors[1] = bound(0.0f, gl_foggreen.value, 1.0f);
		mcolors[2] = bound(0.0f, gl_fogblue.value, 1.0f); 
		mcolors[3] = exp(gl_fogdensity.value / dist);
		break;

	default:
		glDisable(GL_FOG);
		r_viewleaf->worldfogenabled = false;
		mcolors[0] = 0.0f;
		mcolors[1] = 0.0f;
		mcolors[2] = 0.0f;
		mcolors[3] = 0.0f;
		break;
	}

	if (r_viewleaf->worldfogenabled) {
		if (nehahra) {
			glFogi(GL_FOG_MODE, GL_EXP2);
			glFogfv(GL_FOG_COLOR, mcolors); 
			glFogf(GL_FOG_DENSITY, bound(0.0f, gl_fogdensity.value, 1.0f) / 100.0f); 
		} else {
			glFogi(GL_FOG_MODE, GL_LINEAR);
			glFogfv(GL_FOG_COLOR, mcolors); 
			glFogf(GL_FOG_START, gl_fogstart.value); 
			glFogf(GL_FOG_END, gl_fogend.value);
			glFogf(GL_FOG_DENSITY, bound(0.0f, gl_fogdensity.value, 1.0f) / 100.0f); 
		}
	}
}
#endif

/*
=============
V_CalcPowerupCshift
=============
*/
void V_CalcPowerupCshift (void)
{
	float	fraction;

	if (cl.items & IT_QUAD)
	{
		cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 0;
		cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 0;
		cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 255;
		fraction = bound(0, v_quadcshift.value, 1);
		cl.cshifts[CSHIFT_POWERUP].percent = 30 * fraction;
	}
	else if (cl.items & IT_SUIT)
	{
		cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 0;
		cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 255;
		cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 0;
		fraction = bound(0, v_suitcshift.value, 1);
		cl.cshifts[CSHIFT_POWERUP].percent = 20 * fraction;
	}
	else if (cl.items & IT_INVISIBILITY)
	{
		cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 100;
		cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 100;
		cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 100;
		fraction = bound(0, v_ringcshift.value, 1);
		cl.cshifts[CSHIFT_POWERUP].percent = 100 * fraction;
	}
	else if (cl.items & IT_INVULNERABILITY)
	{
		cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 255;
		cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 255;
		cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 0;
		fraction = bound(0, v_pentcshift.value, 1);
		cl.cshifts[CSHIFT_POWERUP].percent = 30 * fraction;
	}
	else
	{
		cl.cshifts[CSHIFT_POWERUP].percent = 0;
	}
}

/*
=============
V_CalcBlend

Mostly Havoc code.
=============
*/
#ifdef	GLQUAKE
void V_CalcBlend (void)
{
	float	r, g, b, a, a2;
	int		j;

	r = 0.0f;
	g = 0.0f;
	b = 0.0f;
	a = 0.0f;

	if (cls.state != ca_connected)
	{
		cl.cshifts[CSHIFT_CONTENTS] = cshift_empty;
		cl.cshifts[CSHIFT_POWERUP].percent = 0;
	}
	else
	{
		V_CalcPowerupCshift ();
	}

	// drop the damage value
	cl.cshifts[CSHIFT_DAMAGE].percent -= host_frametime * 150;
	if (cl.cshifts[CSHIFT_DAMAGE].percent <= 0)
		cl.cshifts[CSHIFT_DAMAGE].percent = 0;

	// drop the bonus value
	cl.cshifts[CSHIFT_BONUS].percent -= host_frametime * 100;
	if (cl.cshifts[CSHIFT_BONUS].percent <= 0)
		cl.cshifts[CSHIFT_BONUS].percent = 0;

	for (j=0 ; j<NUM_CSHIFTS ; j++)	
	{
		a2 = cl.cshifts[j].percent * (1.0f / 255.0f);

		if (!a2) continue;

		if (a2 > 1)
		{
			a2 = 1;
		}
		r += (cl.cshifts[j].destcolor[0]-r) * a2;
		g += (cl.cshifts[j].destcolor[1]-g) * a2;
		b += (cl.cshifts[j].destcolor[2]-b) * a2;
		a = 1 - (1 - a) * (1 - a2); // correct alpha multiply...  took a while to find it on the web
	}

	// saturate color (to avoid blending in black)
	if (a)
	{
		a2 = 1 / a;
		r *= a2;
		g *= a2;
		b *= a2;
	}
	v_blend[0] = bound(0, r/255.0f, 1);
	v_blend[1] = bound(0, g/255.0f, 1);
	v_blend[2] = bound(0, b/255.0f, 1);
	v_blend[3] = bound(0, a, 1);
}

/*
=============
V_AddLightBlend
=============
*/
void V_AddLightBlend (float r, float g, float b, float a2)
{
	float	a;

	if (!gl_polyblend.value || !v_dlightcshift.value) return;

	a2 = a2 * bound(0, v_dlightcshift.value, 1) * 1;

	v_blend[3] = a = v_blend[3] + a2 * (1 - v_blend[3]);

	if (!a)	return;

	a2 = a2 / a;

	v_blend[0] = v_blend[0] * (1 - a2) + r * a2;
	v_blend[1] = v_blend[1] * (1 - a2) + g * a2;
	v_blend[2] = v_blend[2] * (1 - a2) + b * a2;
}
#endif

/*
=============
V_UpdatePalette
=============
*/
#ifdef	GLQUAKE
void V_UpdatePalette (void)
{
	qboolean		newf;
	int				i, j, c;
	static float	prev_blend[4];
	float			a, rgb[3], gamma, contrast;
	static float	old_gamma, old_contrast, old_hwblend;

	newf = false;

	for (i=0 ; i<4 ; i++)
	{
		if (v_blend[i] != prev_blend[i])
		{
			newf = true;
			prev_blend[i] = v_blend[i];
		}
	}
	gamma = bound(0.3, gl_gamma.value, 3);

	if (gamma != old_gamma)
	{
		old_gamma = gamma;
		newf = true;
	}
	contrast = bound(1, gl_contrast.value, 3);

	if (contrast != old_contrast)
	{
		old_contrast = contrast;
		newf = true;
	}

	if (gl_hwblend.value != old_hwblend)
	{
		newf = true;
		old_hwblend = gl_hwblend.value;
	}

	if (!newf) return;

	a = v_blend[3];

	if (!vid_hwgamma_enabled || !gl_hwblend.value)
	{
		a = 0;
	}
	rgb[0] = 255 * v_blend[0] * a;
	rgb[1] = 255 * v_blend[1] * a;
	rgb[2] = 255 * v_blend[2] * a;

	a = 1 - a;

	if (vid_gamma != 1.0)
	{
		contrast = pow (contrast, vid_gamma);
		gamma /= vid_gamma;
	}

	for (i=0 ; i<256 ; i++)
	{
		for (j=0 ; j<3 ; j++)
		{
			// apply blend and contrast
			c = (i*a + rgb[j]) * contrast;
			if (c > 255)
				c = 255;
			// apply gamma
			c = 255 * pow((c + 0.5)/255.5, gamma) + 0.5;
			c = bound(0, c, 255);
			v_gammaramps[j][i] = c << 8;
		}
	}
	VID_SetDeviceGammaRamp ((unsigned short *)v_gammaramps);
}
#else	// !GLQUAKE
void V_UpdatePalette (void)
{
	int				i, j, r, g, b;
	qboolean		newf, force;
	byte			*basepal, *newpal;
	static cshift_t	prev_cshifts[NUM_CSHIFTS];

	if (cls.state != ca_connected)
	{
		cl.cshifts[CSHIFT_CONTENTS] = cshift_empty;
		cl.cshifts[CSHIFT_POWERUP].percent = 0;
	}
	else
	{
		V_CalcPowerupCshift ();
	}
	newf = false;
	
	for (i=0 ; i<NUM_CSHIFTS ; i++)
	{
		if (cl.cshifts[i].percent != prev_cshifts[i].percent)
		{
			newf = true;
			prev_cshifts[i].percent = cl.cshifts[i].percent;
		}

		for (j=0 ; j<3 ; j++)
		{
			if (cl.cshifts[i].destcolor[j] != prev_cshifts[i].destcolor[j])
			{
				newf = true;
				prev_cshifts[i].destcolor[j] = cl.cshifts[i].destcolor[j];
			}
		}
	}
	
	// drop the damage value
	cl.cshifts[CSHIFT_DAMAGE].percent -= host_frametime * 150;
	if (cl.cshifts[CSHIFT_DAMAGE].percent <= 0)
		cl.cshifts[CSHIFT_DAMAGE].percent = 0;

	// drop the bonus value
	cl.cshifts[CSHIFT_BONUS].percent -= host_frametime * 100;
	if (cl.cshifts[CSHIFT_BONUS].percent <= 0)
		cl.cshifts[CSHIFT_BONUS].percent = 0;

	force = V_CheckGamma ();
	if (!newf && !force) return;
			
	basepal = host_basepal;
	newpal = current_pal;	// Tonik: so we can use current_pal for screenshots
	
	for (i=0 ; i<256 ; i++)
	{
		r = basepal[0];
		g = basepal[1];
		b = basepal[2];
		basepal += 3;
	
		for (j=0 ; j<NUM_CSHIFTS ; j++)	
		{
			r += (cl.cshifts[j].percent*(cl.cshifts[j].destcolor[0]-r)) >> 8;
			g += (cl.cshifts[j].percent*(cl.cshifts[j].destcolor[1]-g)) >> 8;
			b += (cl.cshifts[j].percent*(cl.cshifts[j].destcolor[2]-b)) >> 8;
		}
		
		newpal[0] = gammatable[r];
		newpal[1] = gammatable[g];
		newpal[2] = gammatable[b];
		newpal += 3;
	}
	VID_ShiftPalette (current_pal);
}
#endif	// !GLQUAKE


/* 
============================================================================== 
 
				VIEW RENDERING 
 
============================================================================== 
*/ 

/*
==================
CalcGunAngle
==================
*/
void CalcGunAngle (void)
{	
	float	yaw, pitch, move;
	static float oldyaw = 0;
	static float oldpitch = 0;
	
	yaw = r_refdef.viewangles[YAW];
	pitch = -r_refdef.viewangles[PITCH];

	yaw = AngleDelta(yaw - r_refdef.viewangles[YAW]) * 0.4;
	yaw = bound(-10, yaw, 10);

	pitch = AngleDelta(-pitch - r_refdef.viewangles[PITCH]) * 0.4;
	pitch = bound(-10, pitch, 10);

	move = host_frametime*20;
	if (yaw > oldyaw)
	{
		if (oldyaw + move < yaw)
			yaw = oldyaw + move;
	}
	else
	{
		if (oldyaw - move > yaw)
			yaw = oldyaw - move;
	}
	
	if (pitch > oldpitch)
	{
		if (oldpitch + move < pitch)
			pitch = oldpitch + move;
	}
	else
	{
		if (oldpitch - move > pitch)
			pitch = oldpitch - move;
	}
	
	oldyaw = yaw;
	oldpitch = pitch;

	// joe: this makes it fix when strafing
	cl.viewent.angles[YAW] = r_refdef.viewangles[YAW] + yaw;
	cl.viewent.angles[PITCH] = -(r_refdef.viewangles[PITCH] + pitch);
	cl.viewent.angles[ROLL] = r_refdef.viewangles[ROLL];

	cl.viewent.angles[ROLL] -= v_idlescale.value * sin(cl.time*v_iroll_cycle.value) * v_iroll_level.value;
	cl.viewent.angles[PITCH] -= v_idlescale.value * sin(cl.time*v_ipitch_cycle.value) * v_ipitch_level.value;
	cl.viewent.angles[YAW] -= v_idlescale.value * sin(cl.time*v_iyaw_cycle.value) * v_iyaw_level.value;
}

/*
==============
V_BoundOffsets
==============
*/
void V_BoundOffsets (void)
{
	entity_t	*ent;
	
	ent = &cl_entities[cl.viewentity];

// absolutely bound refresh reletive to entity clipping hull
// so the view can never be inside a solid wall
	r_refdef.vieworg[0] = max(r_refdef.vieworg[0], ent->origin[0] - 14);
	r_refdef.vieworg[0] = min(r_refdef.vieworg[0], ent->origin[0] + 14);
	r_refdef.vieworg[1] = max(r_refdef.vieworg[1], ent->origin[1] - 14);
	r_refdef.vieworg[1] = min(r_refdef.vieworg[1], ent->origin[1] + 14);
	r_refdef.vieworg[2] = max(r_refdef.vieworg[2], ent->origin[2] - 22);
	r_refdef.vieworg[2] = min(r_refdef.vieworg[2], ent->origin[2] + 30);
}

/*
==============
V_AddIdle

Idle swaying
==============
*/
void V_AddIdle (void)
{
	r_refdef.viewangles[ROLL] += v_idlescale.value * sin(cl.time*v_iroll_cycle.value) * v_iroll_level.value;
	r_refdef.viewangles[PITCH] += v_idlescale.value * sin(cl.time*v_ipitch_cycle.value) * v_ipitch_level.value;
	r_refdef.viewangles[YAW] += v_idlescale.value * sin(cl.time*v_iyaw_cycle.value) * v_iyaw_level.value;
}


/*
==============
V_CalcViewRoll

Roll is induced by movement and damage
==============
*/
void V_CalcViewRoll (void)
{
	float		side;
		
	side = V_CalcRoll (cl_entities[cl.viewentity].angles, cl.velocity);
	r_refdef.viewangles[ROLL] += side;

	if (v_dmg_time > 0)
	{
		r_refdef.viewangles[ROLL] += v_dmg_time / v_kicktime.value * v_dmg_roll;
		r_refdef.viewangles[PITCH] += v_dmg_time / v_kicktime.value * v_dmg_pitch;
		v_dmg_time -= host_frametime;
	}

	if (cl.stats[STAT_HEALTH] <= 0)
	{
		r_refdef.viewangles[ROLL] = 80;	// dead view angle
		return;
	}
}


/*
==================
V_CalcIntermissionRefdef

==================
*/
void V_CalcIntermissionRefdef (void)
{
	entity_t	*ent, *view;
	float		old;

// ent is the player model (visible when out of body)
	ent = &cl_entities[cl.viewentity];
// view is the weapon model (only visible from inside body)
	view = &cl.viewent;

	VectorCopy (ent->origin, r_refdef.vieworg);
	VectorCopy (ent->angles, r_refdef.viewangles);
	view->model = NULL;

// always idle in intermission
	old = v_idlescale.value;
	v_idlescale.value = 1;
	V_AddIdle ();
	v_idlescale.value = old;
}

/*
==================
V_CalcRefdef

==================
*/
void V_CalcRefdef (void)
{
	entity_t		*ent, *view;
	int				i;
	vec3_t			forward, right, up, angles;
	float			bob;
	static	float	oldz = 0;

	V_DriftPitch ();

// ent is the player model (visible when out of body)
	ent = &cl_entities[cl.viewentity];

// view is the weapon model (only visible from inside body)
	view = &cl.viewent;
	
// transform the view offset by the model's matrix to get the offset from
// model origin for the view
	ent->angles[YAW] = cl.viewangles[YAW];		// the model should face the view dir
	ent->angles[PITCH] = -cl.viewangles[PITCH];	//
										
	bob = V_CalcBob ();
	
// refresh position
	VectorCopy (ent->origin, r_refdef.vieworg);
	r_refdef.vieworg[2] += cl.viewheight + bob;

// never let it sit exactly on a node line, because a water plane can
// dissapear when viewed with the eye exactly on it.
// the server protocol only specifies to 1/16 pixel, so add 1/32 in each axis
	r_refdef.vieworg[0] += 1.0 / 32;
	r_refdef.vieworg[1] += 1.0 / 32;
	r_refdef.vieworg[2] += 1.0 / 32;

	VectorCopy (cl.viewangles, r_refdef.viewangles);
	V_CalcViewRoll ();
	V_AddIdle ();

// offsets
	angles[PITCH] = -ent->angles[PITCH];	// because entity pitches are
						// actually backward
	angles[YAW] = ent->angles[YAW];
	angles[ROLL] = ent->angles[ROLL];

	AngleVectors (angles, forward, right, up);

	for (i=0 ; i<3 ; i++)
		r_refdef.vieworg[i] += scr_ofsx.value*forward[i] + scr_ofsy.value*right[i] + scr_ofsz.value*up[i];
	
	V_BoundOffsets ();
		
// set up gun position
	VectorCopy (cl.viewangles, view->angles);
	
	CalcGunAngle ();

	VectorCopy (ent->origin, view->origin);
	view->origin[2] += cl.viewheight;

	VectorMA (view->origin, bob*0.4, forward, view->origin);
	view->origin[2] += bob;

// fudge position around to keep amount of weapon visible
// roughly equal with different FOV
	if (scr_viewsize.value == 110)
		view->origin[2] += 1;
	else if (scr_viewsize.value == 100)
		view->origin[2] += 2;
	else if (scr_viewsize.value == 90)
		view->origin[2] += 1;
	else if (scr_viewsize.value == 80)
		view->origin[2] += 0.5;

	view->model = cl.model_precache[cl.stats[STAT_WEAPON]];
	view->frame = cl.stats[STAT_WEAPONFRAME];
	view->colormap = vid.colormap;

// set up the refresh position
	if (v_gunkick.value)	// joe	-- ANTI JARRING!!!
		VectorAdd (r_refdef.viewangles, cl.punchangle, r_refdef.viewangles);

// smooth out stair step ups
	if (cl.onground && ent->origin[2] - oldz > 0)
	{
		oldz += host_frametime * 80;	//fabs(cl.ctime - cl.oldtime);
		if (oldz > ent->origin[2])
			oldz = ent->origin[2];
		if (ent->origin[2] - oldz > 12)
			oldz = ent->origin[2] - 12;
		r_refdef.vieworg[2] += oldz - ent->origin[2];
		view->origin[2] += oldz - ent->origin[2];
	}
	else
	{
		oldz = ent->origin[2];
	}

	if (chase_active.value)
		Chase_Update ();
}

/*
==============
SCR_DrawLocalTime
==============
*/
void SCR_DrawLocalTime (void)
{
	time_t		ltime;
	char		str[80], *format;

	if (!scr_clock.value)
		return;

	format = ((int)scr_clock.value == 2) ? "%H:%M:%S" : 
		((int)scr_clock.value == 3) ? "%a %b %I:%M:%S %p" : 
		((int)scr_clock.value == 4) ? "%a %b %H:%M:%S" : "%I:%M:%S %p";

	time (&ltime);
	strftime (str, sizeof(str)-1, format, localtime(&ltime));

	if (scr_clock_y.value < 0)
		Draw_String (8*scr_clock_x.value, vid.height - sb_lines + 8*scr_clock_y.value, str);
	else
		Draw_String (8*scr_clock_x.value, 8*scr_clock_y.value, str);
}

/*
==============
SCR_DrawFPS
==============
*/
// joe: from FuhQuake
void SCR_DrawFPS (void)
{
	static	float	lastframetime;
	float			t;
	extern	int		fps_count;
	static	float	lastfps;
	int				x, y;
	char			st[80];

	if (!show_fps.value)
		return;

	t = Sys_DoubleTime ();
	if ((t - lastframetime) >= 1.0)
	{
		lastfps = fps_count / (t - lastframetime);
		fps_count = 0;
		lastframetime = t;
	}

	sprintf (st, "%3.1f FPS", lastfps + 0.05);
	x = vid.width - strlen(st) * 8 - 8;
	y = vid.height - sb_lines - 8;

	Draw_String (x, y, st);
}

/*
==============
SCR_DrawSpeed
==============
*/
// joe: copied from [sons]Quake
void SCR_DrawSpeed (void)
{
	int		x, y;
	char		st[8];
	vec3_t		vel;
	float		speed, vspeed, speedunits;
	static	float	maxspeed = 0;
	static	float	display_speed = -1;
	static	double	lastrealtime = 0;

	if (!show_speed.value)
		return;

	if (lastrealtime > realtime)
	{
		lastrealtime = 0;
		display_speed = -1;
		maxspeed = 0;
	}

	VectorCopy (cl.velocity, vel);
	vspeed = vel[2];
	vel[2] = 0;
	speed = VectorLength (vel);

	if (speed > maxspeed)
		maxspeed = speed;

	if (display_speed >= 0)
	{
		sprintf (st, "%3d", (int)display_speed);

		x = vid.width/2 - 80;

		if (scr_viewsize.value >= 120)
			y = vid.height - 16;

		if (scr_viewsize.value < 120)
			y = vid.height - 8*5;

		if (scr_viewsize.value < 110)
			y = vid.height - 8*8;

		if (cl.intermission)
			y = vid.height - 16;

		Draw_Fill (x, y-1, 160, 1, 10);
		Draw_Fill (x, y+9, 160, 1, 10);
		Draw_Fill (x+32, y-2, 1, 13, 10);
		Draw_Fill (x+64, y-2, 1, 13, 10);
		Draw_Fill (x+96, y-2, 1, 13, 10);
		Draw_Fill (x+128, y-2, 1, 13, 10);

		Draw_Fill (x, y, 160, 9, 52);

		speedunits = display_speed;
		if (display_speed <= 500)
			Draw_Fill (x, y, (int)(display_speed/3.125), 9, 100);
		else 
		{   
			while (speedunits > 500)
				speedunits -= 500;
			Draw_Fill (x, y, (int)(speedunits/3.125), 9, 68);
		}
		Draw_String (x + 36 - strlen(st) * 8, y, st);
	}

	if (realtime - lastrealtime >= 0.1)
	{
		lastrealtime = realtime;
		display_speed = maxspeed;
		maxspeed = 0;
	}
}

/*
==================
V_RenderView

The player's clipping box goes from (-16 -16 -24) to (16 16 32) from
the entity origin, so any view position inside that will be valid
==================
*/
void V_RenderView (void)
{
	if (con_forcedup)
		return;

// don't allow cheats in multiplayer
	if (cl.maxclients > 1)
	{
		Cvar_Set ("scr_ofsx", "0");
		Cvar_Set ("scr_ofsy", "0");
		Cvar_Set ("scr_ofsz", "0");
	}

	if (cls.state != ca_connected)
	{
#ifdef GLQUAKE
		V_CalcBlend ();
#endif
		return;
	}

	if (cl.intermission)
	{	// intermission / finale rendering
		V_CalcIntermissionRefdef ();	
	}
	else
	{
		if (!cl.paused)
			V_CalcRefdef ();
	}

	R_PushDlights ();

	if (lcd_x.value)
	{
		// render two interleaved views
		int	i;

		vid.rowbytes <<= 1;
		vid.aspect *= 0.5;

		r_refdef.viewangles[YAW] -= lcd_yaw.value;
		for (i=0 ; i<3 ; i++)
			r_refdef.vieworg[i] -= right[i] * lcd_x.value;
		R_RenderView ();

		vid.buffer += vid.rowbytes >> 1;

		R_PushDlights ();

		r_refdef.viewangles[YAW] += lcd_yaw.value * 2;
		for (i=0 ; i<3 ; i++)
			r_refdef.vieworg[i] += 2 * right[i] * lcd_x.value;
		R_RenderView ();

		vid.buffer -= vid.rowbytes >> 1;

		r_refdef.vrect.height <<= 1;

		vid.rowbytes >>= 1;
		vid.aspect *= 2;
	}
	else
	{
		R_RenderView ();
	}
}

//============================================================================

/*
=============
V_Init
=============
*/
void V_Init (void)
{
	Cmd_AddCommand ("v_cshift", V_cshift_f);
	Cmd_AddCommand ("bf", V_BonusFlash_f);
	Cmd_AddCommand ("centerview", V_StartPitchDrift);

	Cvar_RegisterVariable (&lcd_x);
	Cvar_RegisterVariable (&lcd_yaw);

	Cvar_RegisterVariable (&v_centermove);
	Cvar_RegisterVariable (&v_centerspeed);

	Cvar_RegisterVariable (&v_iyaw_cycle);
	Cvar_RegisterVariable (&v_iroll_cycle);
	Cvar_RegisterVariable (&v_ipitch_cycle);
	Cvar_RegisterVariable (&v_iyaw_level);
	Cvar_RegisterVariable (&v_iroll_level);
	Cvar_RegisterVariable (&v_ipitch_level);
	Cvar_RegisterVariable (&v_idlescale);

	Cvar_RegisterVariable (&crosshair);
	Cvar_RegisterVariable (&crosshaircolor);
	Cvar_RegisterVariable (&crosshairsize);
	Cvar_RegisterVariable (&cl_crossx);
	Cvar_RegisterVariable (&cl_crossy);

	Cvar_RegisterVariable (&scr_ofsx);
	Cvar_RegisterVariable (&scr_ofsy);
	Cvar_RegisterVariable (&scr_ofsz);

	Cvar_RegisterVariable (&scr_clock);
	Cvar_RegisterVariable (&scr_clock_x);
	Cvar_RegisterVariable (&scr_clock_y);
	Cvar_RegisterVariable (&show_speed);
	Cvar_RegisterVariable (&show_fps);

	Cvar_RegisterVariable (&cl_rollspeed);
	Cvar_RegisterVariable (&cl_rollangle);
	Cvar_RegisterVariable (&cl_bob);
	Cvar_RegisterVariable (&cl_bobcycle);
	Cvar_RegisterVariable (&cl_bobup);

	Cvar_RegisterVariable (&v_kicktime);
	Cvar_RegisterVariable (&v_kickroll);
	Cvar_RegisterVariable (&v_kickpitch);	
	
	Cvar_RegisterVariable (&v_gunkick);

	Cvar_RegisterVariable (&v_bonusflash);
	Cvar_RegisterVariable (&v_damagecshift);
	Cvar_RegisterVariable (&v_quadcshift);
	Cvar_RegisterVariable (&v_suitcshift);
	Cvar_RegisterVariable (&v_ringcshift);
	Cvar_RegisterVariable (&v_pentcshift);
#ifdef GLQUAKE
	Cvar_RegisterVariable (&v_dlightcshift);
	Cvar_RegisterVariable (&gl_gamma);
	Cvar_RegisterVariable (&gl_contrast);
	Cvar_RegisterVariable (&gl_hwblend);
#endif

	Cvar_RegisterVariable (&v_gamma);
	Cvar_RegisterVariable (&v_contrast);

#ifndef GLQUAKE
	V_BuildGammaTable (v_gamma.value, v_contrast.value);
#endif
}

